//-----------------------------------------------------------------------
// Copyright (c) 2020 Snap Inc.
//-----------------------------------------------------------------------

#include <required.glsl>

uniform sampler2D baseTex;
uniform mat3      baseTexTransform;

varying vec2 varTex0;
varying vec2 varTex1;

#include <envProvider/includes/studioUtils.glsl>
#include <envProvider/includes/envProviderUtils.glsl>

#ifdef VERTEX_SHADER
attribute vec4 position;
attribute vec2 texture0;

void main(){
    varTex0=vec2(baseTexTransform*vec3(texture0,1.0));
    varTex1 = texture0;

    gl_Position = vec4(sign(position.xy), 0.0, 1.0);
}
#endif // VERTEX_SHADER

#ifdef FRAGMENT_SHADER

float calcHDR (float t) {
    const float e = 2.718;

    float t042 = t * 0.65;
    float h = pow(e, t * 3.0) / 10.0;
    float t065 = 0.19 + max(pow(h, 2.3), 0.0); 
    float t082 = 0.34 + max(pow(h, 4.0), 0.0);
    float t100 = -0.27 + max(pow(h, 5.8), 0.0);

    float flag0 = step(t, 0.42);
    float flag1 = step(t, 0.65);
    float flag2 = step(t, 0.82);

    return mix(mix(mix(t100, t082, flag2), t065, flag1), t042, flag0);
}

vec3 inverseACESToneMapping(vec3 LDR) {
    float LDR_Luma = 1e-5;

#ifdef GL_FRAGMENT_PRECISION_HIGH
    vec3 LDR_HSL = RGBToHSL(LDR);
    LDR_Luma = LDR_HSL.z;
#else
    LDR_Luma = luma(LDR);
#endif

    float LDR_val = ensureIsNotZero(LDR_Luma);
    float HDR_val = calcHDR(LDR_val);

    return LDR * HDR_val/LDR_val;
}

void main(){
    vec4 reprojColor = texture2D(baseTex, varTex0);

    vec4 finColor = reprojColor;
    finColor = vec4(srgbToLinear(finColor.rgb), finColor.a);
    finColor = vec4(inverseACESToneMapping(finColor.rgb), finColor.a);

    // vec3(0.85, 0.87, 0.91) to HDR
    vec3 skyColor = vec3(5.279, 5.403, 5.652);
    float skyFactor = clamp(varTex1.y * 2.0 - 1.0, 0.0, 1.0);
    // 0.0 .. 0.1 -- hack to make dim scenes brighter by faking the sky above the object reflections
    skyFactor = 0.1 * sin(skyFactor*PI/2.0);

    finColor.rgb = mix(finColor.rgb, skyColor, skyFactor);

    finColor = encodeRGBD(finColor.rgb);

    gl_FragColor = finColor;
}
#endif // FRAGMENT_SHADER
