//-----------------------------------------------------------------------
// Copyright (c) 2017 Snap Inc.
//-----------------------------------------------------------------------

#ifdef SC_TEXTURE_EXTERNAL
#extension GL_OES_EGL_image_external : require
#endif // #ifdef SC_TEXTURE_EXTERNAL

#define SC_USE_USER_DEFINED_VS_MAIN
#define STD_DISABLE_VERTEX_NORMAL
#define STD_DISABLE_VERTEX_TANGENT
#define STD_DISABLE_VERTEX_TEXTURE1

#include <std_vs.glsl>
#include <std_fs.glsl>

#ifdef SC_TEXTURE_EXTERNAL
DECLARE_TEXTURE_SAMPLER_TYPE(screenTexture, samplerExternalOES)
#endif // #ifdef SC_TEXTURE_EXTERNAL

DECLARE_TEXTURE_SAMPLER_UNIFORMS(screenTexture)
uniform mat3 screenTextureTransform;

#ifdef USE_MESH_TRANSFORM
uniform mat3 meshTransform;
#endif // #ifdef USE_MESH_TRANSFORM

#ifdef VERTEX_SHADER

void main(void) {
    sc_Vertex_t v = sc_LoadVertexAttributes();

    #ifdef ONLY_VERTEX_ATTRIBUTE
    vec2 texture0 = (v.position.xy + vec2(1.0)) * 0.5;
    #else // #ifdef ONLY_VERTEX_ATTRIBUTE
    vec2 texture0 = v.texture0.xy;
    #endif // #else // #ifdef ONLY_VERTEX_ATTRIBUTE

    v.texture0.xy = vec2(screenTextureTransform * vec3(texture0, 1.0));

    #ifdef USE_MESH_TRANSFORM
    v.position.xy = vec2(meshTransform * vec3(position.xy, 1.0));
    #endif // #ifdef USE_MESH_TRANSFORM

    sc_ProcessVertex(v);
}

#endif // #ifdef VERTEX_SHADER

#ifdef FRAGMENT_SHADER

void main(void) {

#ifdef sc_EnableFeatureLevelES3
    vec4 screenTextureColor = screenTextureSampleViewLevel(varTex0, screenTextureLevel);
#else // #ifdef sc_EnableFeatureLevelES3
    vec4 screenTextureColor = sc_textureExternalOESLod(screenTexture, screenTextureSize.xy, varTex0, screenTextureLevel);
#endif // #else // #ifdef sc_EnableFeatureLevelES3

#ifdef SWAP_R_B_CHANNELS
    gl_FragColor = screenTextureColor.bgra;
#elif defined(GRAYSCALE_AS_ALPHA)
    gl_FragColor = vec4(1.0, 1.0, 1.0, screenTextureColor.r);
#else
    gl_FragColor = screenTextureColor;
#endif

}

#endif // #ifdef FRAGMENT_SHADER
