//-----------------------------------------------------------------------
// Copyright (c) 2017 Snap Inc.
//-----------------------------------------------------------------------

#pragma once

#include <required.glsl>

struct sc_PointLight_t {
    vec3 position;
    vec4 color;
};

struct sc_DirectionalLight_t {
    vec3 direction;
    vec4 color;
};

struct sc_AmbientLight_t {
    vec3 color;
    float intensity;

#ifndef SC_REMOVE_ENVMAP_FROM_AMBIENT_LIGHT_STRUCT  // 09/06/2018 HACK: Adreno GPUs can't handle textures in arrays of structs, some uniform values get corrupted, so we remove the textures from the ambient struct, since we don't use them anyway. This can be removed once we have no old uber shaders that sample envmaps from the ambient light struct.
    float weight;
    //vec3 position; // will be used for probes location

    // Members below this line are deprecated.

    vec3 sphericalHarmonics[9];

    vec3 rotation;
    float exposure;
    sampler2D diffuseEnvMap; // HACK: make sure textures are at the end of the struct. Google Pixel drivers seem to have a bug where they fail on some uniforms that come after textures in a struct!
    sampler2D specularEnvMap;
#endif
};

// The environment light is specified using loose uniforms to avoid device specific bugs with samplers in arrays/structs.
uniform sampler2D sc_EnvmapDiffuse;
uniform vec4 sc_EnvmapDiffuseSize;

uniform sampler2D sc_EnvmapSpecular;
uniform vec4 sc_EnvmapSpecularSize;

uniform vec3 sc_EnvmapRotation;
uniform float sc_EnvmapExposure;

// In some cases we support diffuse ambient light in SH coefficients. This replaces the diffuse envmap.
uniform vec3 sc_Sh[9];
uniform float sc_ShIntensity;

struct sc_AmbientLightCompatibility_t {
    vec3 color;
};

struct sc_Camera_t {
    // GPU_BUG_010 [LOOK-42347] 04/03/2019 Mali 400: This compiler forces all members of a struct to have the same precision, and assumes all members are used in a shader stage if any one member is used. We always use sc_Camera.position in the fragment shader, which forces precision of all members to mediump. Once we attempt to use sc_Camera.clipPlanes in the vertex shader, it creates highp uniforms for all members, which conflict with the mediump uniforms from the fragment shader. This means that even the "aspect" member conflicts with itself, even though it's not used at all, neither in vs nor fs. The solution is to force all members to fragment precision.
    sc_FragmentPrecision vec3 position;
    sc_FragmentPrecision float aspect;
    sc_FragmentPrecision vec2 clipPlanes; //x - near, y - far
};

uniform vec4 sc_Time;
#define sc_TimeElapsed (sc_Time.x)
#define sc_TimeDelta   (sc_Time.y)

uniform vec4 sc_GeometryInfo;
#define sc_PrimitiveCount (sc_GeometryInfo.x)
#define sc_InstanceCount  (sc_GeometryInfo.y)

#if (sc_StereoRenderingMode == sc_StereoRendering_InstancedClipped) || (sc_StereoRenderingMode == sc_StereoRendering_Multiview)

uniform mat4 sc_ModelViewProjectionMatrixArray[sc_NumStereoViews];
uniform mat4 sc_ModelViewProjectionMatrixInverseArray[sc_NumStereoViews];
uniform mat4 sc_ViewProjectionMatrixArray[sc_NumStereoViews];
uniform mat4 sc_ViewProjectionMatrixInverseArray[sc_NumStereoViews];
uniform mat4 sc_ModelViewMatrixArray[sc_NumStereoViews];
uniform mat4 sc_ModelViewMatrixInverseArray[sc_NumStereoViews];
uniform mat3 sc_ViewNormalMatrixArray[sc_NumStereoViews];
uniform mat3 sc_ViewNormalMatrixInverseArray[sc_NumStereoViews];
uniform mat4 sc_ProjectionMatrixArray[sc_NumStereoViews];
uniform mat4 sc_ProjectionMatrixInverseArray[sc_NumStereoViews];
uniform mat4 sc_ViewMatrixArray[sc_NumStereoViews];
uniform mat4 sc_ViewMatrixInverseArray[sc_NumStereoViews];

/* Since gl_InstanceID/gl_ViewID_OVR won't be available in fragment shaders,
 * varStereoViewID will be assigned in clipPosition() (in std_vs.glsl)
 * and used instead of sc_StereoViewID macro in fragment shaders.
 */
#ifdef sc_EnableFeatureLevelES3
#define VarStereoViewIDType int
flat varying VarStereoViewIDType varStereoViewID;
#else
#define VarStereoViewIDType float
varying VarStereoViewIDType varStereoViewID;
#endif

#if sc_StereoRenderingMode == sc_StereoRendering_InstancedClipped
#if sc_StereoRendering_IsClipDistanceEnabled
// GLES20ForwardRenderer only enables a single clip distance slot.
#define sc_ClipDistance0 gl_ClipDistance[0]
#else
varying float varClipDistance;
#define sc_ClipDistance0 varClipDistance
#endif // sc_StereoRendering_IsClipDistanceEnabled
#endif // sc_StereoRendering_InstancedClipped

#ifdef FRAGMENT_SHADER
#define sc_StereoViewIndex int(varStereoViewID)
#else
#define sc_StereoViewIndex sc_StereoViewID
#endif

#define sc_ModelViewProjectionMatrix        (sc_ModelViewProjectionMatrixArray[sc_StereoViewIndex])
#define sc_ModelViewProjectionMatrixInverse (sc_ModelViewProjectionMatrixInverseArray[sc_StereoViewIndex])
#define sc_ViewProjectionMatrix             (sc_ViewProjectionMatrixArray[sc_StereoViewIndex])
#define sc_ViewProjectionMatrixInverse      (sc_ViewProjectionMatrixInverseArray[sc_StereoViewIndex])
#define sc_ModelViewMatrix                  (sc_ModelViewMatrixArray[sc_StereoViewIndex])
#define sc_ModelViewMatrixInverse           (sc_ModelViewMatrixInverseArray[sc_StereoViewIndex])
#define sc_ViewNormalMatrix                 (sc_ViewNormalMatrixArray[sc_StereoViewIndex])
#define sc_ViewNormalMatrixInverse          (sc_ViewNormalMatrixInverseArray[sc_StereoViewIndex])
#define sc_ProjectionMatrix                 (sc_ProjectionMatrixArray[sc_StereoViewIndex])
#define sc_ProjectionMatrixInverse          (sc_ProjectionMatrixInverseArray[sc_StereoViewIndex])
#define sc_ViewMatrix                       (sc_ViewMatrixArray[sc_StereoViewIndex])
#define sc_ViewMatrixInverse                (sc_ViewMatrixInverseArray[sc_StereoViewIndex])

#else

uniform mat4 sc_ModelViewProjectionMatrix;
uniform mat4 sc_ModelViewProjectionMatrixInverse;
uniform mat4 sc_ViewProjectionMatrix;
uniform mat4 sc_ViewProjectionMatrixInverse;
uniform mat4 sc_ModelViewMatrix;
uniform mat4 sc_ModelViewMatrixInverse;
uniform mat3 sc_ViewNormalMatrix;
uniform mat3 sc_ViewNormalMatrixInverse;
uniform mat4 sc_ProjectionMatrix;
uniform mat4 sc_ProjectionMatrixInverse;
uniform mat4 sc_ViewMatrix;
uniform mat4 sc_ViewMatrixInverse;

#endif

float sc_GetStereoViewIndex() {
    #ifdef sc_StereoViewIndex
    return float(sc_StereoViewIndex);
    #else
    return 0.0;
    #endif
}

uniform mat4 sc_ModelMatrix;
uniform mat4 sc_ModelMatrixInverse;

uniform mat3 sc_NormalMatrix;
uniform mat3 sc_NormalMatrixInverse;

uniform vec4 sc_WindowToViewportTransform; // for a viewport given by (x, y, width, height) this uniform will be (1/width, 1/height, -x/width, -y/height)

uniform sc_Camera_t sc_Camera;

#ifdef sc_PointLightsCount
uniform sc_PointLight_t sc_PointLights[sc_PointLightsCount];
#endif

#ifdef sc_DirectionalLightsCount
uniform sc_DirectionalLight_t sc_DirectionalLights[sc_DirectionalLightsCount];
#endif

#ifdef sc_AmbientLightsCount
uniform sc_AmbientLight_t sc_AmbientLights[sc_AmbientLightsCount];
#endif

// backward compatibility
sc_AmbientLightCompatibility_t getAmbientLightCompatibility() {
    sc_AmbientLightCompatibility_t ambient;
#ifdef sc_AmbientLightsCount
    ambient.color = sc_AmbientLights[0].color * sc_AmbientLights[0].intensity;
#else
    ambient.color = vec3(0.0);
#endif
    return ambient;
}
#define sc_AmbientLight (getAmbientLightCompatibility())

//
// Texture rendering layouts
//
// IRenderer.h
// enum class TextureRenderingLayout { Regular, InstancedClipped, Multiview };
//

#ifndef sc_TextureRenderingLayout_Regular

    #define sc_TextureRenderingLayout_Regular 0
    #define sc_TextureRenderingLayout_StereoInstancedClipped 1
    #define sc_TextureRenderingLayout_StereoMultiview 2

#endif // #ifndef sc_TextureRenderingLayout_Regular

//
// sc_GetViewSamplingCoords
//

vec3 sc_GetViewSamplingCoords(float renderingLayout, float viewIndex, vec2 uv, float pixelHeight) {
    if (renderingLayout == float(sc_TextureRenderingLayout_Regular)) {
        return vec3(uv, 0.0);
    } else if (renderingLayout == float(sc_TextureRenderingLayout_StereoInstancedClipped)) {
        vec3 uvi = vec3(uv, 0.0);

        #if sc_StereoRenderingMode == sc_StereoRendering_Multiview
        uvi.y = (uvi.y + viewIndex) * 0.5;
        uvi.y = (viewIndex < 0.5)
            ? min(uvi.y, 0.5 - pixelHeight * 0.5)
            : max(uvi.y, 0.5 + pixelHeight * 0.5);
        #endif

        return uvi;
    } else { // sc_TextureRenderingLayout_StereoMultiview
        vec3 uvi = vec3(uv, viewIndex);
        return uvi;
    }
}

//
// sc_SampleView + sc_SampleViewLevel for sampler2D and sampler2DArray
//

#ifdef FRAGMENT_SHADER
vec4 sc_SampleView(sc_FragmentPrecision sampler2D sampler, float renderingLayout, float viewIndex, vec2 uv, float pixelHeight) {
    vec3 uvi = sc_GetViewSamplingCoords(renderingLayout, viewIndex, uv, pixelHeight);
    return texture2D(sampler, uvi.xy);
}
#endif // #ifdef FRAGMENT_SHADER

#ifdef sc_CanUseTextureLod
vec4 sc_SampleViewLevel(sc_FragmentPrecision sampler2D sampler, float renderingLayout, float viewIndex, vec2 uv, float pixelHeight, float level) {
    vec3 uvi = sc_GetViewSamplingCoords(renderingLayout, viewIndex, uv, pixelHeight);
    return texture2DLod(sampler, uvi.xy, level);
}
#endif

#ifdef sc_EnableFeatureLevelES3

#ifdef FRAGMENT_SHADER
vec4 sc_SampleView(sc_FragmentPrecision sampler2DArray sampler, float renderingLayout, float viewIndex, vec2 uv, float pixelHeight) {
    vec3 uvi = sc_GetViewSamplingCoords(renderingLayout, viewIndex, uv, pixelHeight);
    return texture2D(sampler, uvi);
}
#endif // #ifdef FRAGMENT_SHADER

#ifdef sc_CanUseTextureLod
vec4 sc_SampleViewLevel(sc_FragmentPrecision sampler2DArray sampler, float renderingLayout, float viewIndex, vec2 uv, float pixelHeight, float level) {
    vec3 uvi = sc_GetViewSamplingCoords(renderingLayout, viewIndex, uv, pixelHeight);
    return texture2DLod(sampler, uvi, level);
}
#endif

#endif // #ifdef sc_EnableFeatureLevelES3

//
// Decide render target sampler's layout based on definitions from GLShaderSourceBuilder.
// Fallback to "Regular=0".
//

#macro DECIDE_TEXTURE_RENDERING_LAYOUT(SAMPLER_NAME)

    #ifndef SAMPLER_NAME##Layout
        #ifdef sc_EnableInstancedClippedStereoRendering
            #define SAMPLER_NAME##Layout sc_TextureRenderingLayout_StereoInstancedClipped
        #endif // #ifdef sc_EnableInstancedClippedStereoRendering
        #ifdef sc_EnableMultiviewStereoRendering
            #define SAMPLER_NAME##Layout sc_TextureRenderingLayout_StereoMultiview
        #endif // #ifdef sc_EnableMultiviewStereoRendering
    #endif // #ifndef SAMPLER_NAME##Layout
    #ifndef SAMPLER_NAME##Layout
        #define SAMPLER_NAME##Layout sc_TextureRenderingLayout_Regular
    #endif // #ifndef SAMPLER_NAME##Layout

#endmacro // #macro DECIDE_TEXTURE_RENDERING_LAYOUT(SAMPLER_NAME)

//
// Decide render target sampler's type based on layout.
// Fallback to "sampler2D" and "sc_SamplerTypeIndex_Undefined"
//

#ifndef sc_SamplerTypeIndex_sampler2D
    #define sc_SamplerTypeIndex_Undefined 0
    #define sc_SamplerTypeIndex_sampler2D 1
    #define sc_SamplerTypeIndex_sampler2DArray 2
    #define sc_SamplerTypeIndex_samplerExternalOES 3
    #define sc_SamplerTypeIndex_sampler3D 4
    #define sc_SamplerTypeIndex_samplerCube 5
#endif

#macro DECLARE_TEXTURE_SAMPLER_TYPE(SAMPLER_NAME, SAMPLER_TYPE)
    #define SAMPLER_NAME##SamplerType SAMPLER_TYPE
    #define SAMPLER_NAME##SamplerTypeIndex sc_SamplerTypeIndex_##SAMPLER_TYPE
#endmacro

#macro DECIDE_TEXTURE_SAMPLER_TYPE(SAMPLER_NAME)

    #ifndef SAMPLER_NAME##SamplerType
        #ifdef SAMPLER_NAME##Layout
            #if SAMPLER_NAME##Layout == sc_TextureRenderingLayout_StereoMultiview
                DECLARE_TEXTURE_SAMPLER_TYPE(SAMPLER_NAME, sampler2DArray)
            #endif // #else // #if SAMPLER_NAME##Layout == sc_TextureRenderingLayout_StereoMultiview
        #endif // #ifdef SAMPLER_NAME##Layout
    #endif // #ifndef SAMPLER_NAME##Sampler
    #ifndef SAMPLER_NAME##SamplerType
        DECLARE_TEXTURE_SAMPLER_TYPE(SAMPLER_NAME, sampler2D)
    #endif // #ifndef SAMPLER_NAME##SamplerType
    #ifndef SAMPLER_NAME##SamplerTypeIndex
        SAMPLER_NAME##SamplerTypeIndex sc_SamplerTypeIndex_Undefined
    #endif // #ifndef SAMPLER_NAME##SamplerTypeIndex

#endmacro // #macro DECIDE_TEXTURE_SAMPLER_TYPE(SAMPLER_NAME)

#macro DECLARE_TEXTURE_UNIFORM_ACCESSOR(NAME, UNIFORM_SELECTOR)

    #ifndef NAME
        #define NAME UNIFORM_SELECTOR
        #define NAME##Runtime 1
    #else // #ifndef SAMPLER_NAME##Layout
        #define NAME##Runtime 0
    #endif // #else // #ifndef SAMPLER_NAME##Layout

#endmacro // #macro DECLARE_TEXTURE_UNIFORM_ACCESSOR(NAME, UNIFORM_SELECTOR)

#macro DECLARE_TEXTURE_SAMPLER_UNIFORMS(SAMPLER_NAME)

    DECIDE_TEXTURE_RENDERING_LAYOUT(SAMPLER_NAME)
    DECIDE_TEXTURE_SAMPLER_TYPE(SAMPLER_NAME)

    //
    // Declare the typed sampler with its properties and accessors:
    //   > dimentsions
    //   > dimentsions reciprocal
    //   > sampler view (slice)
    //

    uniform sc_FragmentPrecision SAMPLER_NAME##SamplerType SAMPLER_NAME;

    // SC_ENABLE_LEGACY_TEXTURE_SAMPLER_UNIFORMS
    // - Uses legacy texture uniforms: Size.
    // - Emulates new texture uniforms: Dims, DimsRecip, View.

    #if SC_ENABLE_LEGACY_TEXTURE_SAMPLER_UNIFORMS

        // Defines Size, Dims and DimsRecip (compat macros with default ones),
        // restricts View to compile-time zeros. This scope exists because we
        // support pretty old and buggy Adreno GPUs (305) and Android OSs (4.4).

        uniform vec4 SAMPLER_NAME##Size; // x=width, y=height, z=1/width, w=1/height (Deprecated)
        #define SAMPLER_NAME##Dims      (vec4(SAMPLER_NAME##Size.xy, 1.0, 1.0))
        #define SAMPLER_NAME##DimsRecip (vec4(SAMPLER_NAME##Size.zw, 1.0, 1.0))
        #define SAMPLER_NAME##View      (vec4(0.0))
        #define SAMPLER_NAME##Layer     0.0
        #define SAMPLER_NAME##Level     0.0

    #else // #if SC_ENABLE_LEGACY_TEXTURE_SAMPLER_UNIFORMS

        //
        // Defines Dims, DimsRecip, Size (legacy), allows runtime View.
        //

        uniform vec4 SAMPLER_NAME##Dims;      // x=width, y=height, z=layers/depth, w=levels
        uniform vec4 SAMPLER_NAME##DimsRecip; // = 1.0 / Dims
        #define SAMPLER_NAME##Size (vec4(SAMPLER_NAME##Dims.xy, SAMPLER_NAME##DimsRecip.xy))

    #endif // #else // #if SC_ENABLE_LEGACY_TEXTURE_SAMPLER_UNIFORMS

    #define SAMPLER_NAME##Width       SAMPLER_NAME##Dims.x
    #define SAMPLER_NAME##Height      SAMPLER_NAME##Dims.y
    #define SAMPLER_NAME##Layers      SAMPLER_NAME##Dims.z
    #define SAMPLER_NAME##Depth       SAMPLER_NAME##Dims.z
    #define SAMPLER_NAME##Levels      SAMPLER_NAME##Dims.w
    #define SAMPLER_NAME##PixelWidth  SAMPLER_NAME##DimsRecip.x
    #define SAMPLER_NAME##PixelHeight SAMPLER_NAME##DimsRecip.y
    #define SAMPLER_NAME##PixelDepth  SAMPLER_NAME##DimsRecip.z

    //
    // View properties can be defined at compile time.
    //

    DECLARE_TEXTURE_UNIFORM_ACCESSOR(SAMPLER_NAME##Layer, SAMPLER_NAME##View.x)
    DECLARE_TEXTURE_UNIFORM_ACCESSOR(SAMPLER_NAME##Level, SAMPLER_NAME##View.y)
    DECLARE_TEXTURE_UNIFORM_ACCESSOR(SAMPLER_NAME##Layout, SAMPLER_NAME##View.z)

    #if SAMPLER_NAME##LayerRuntime || SAMPLER_NAME##LevelRuntime || SAMPLER_NAME##LayoutRuntime
    uniform vec4 SAMPLER_NAME##View; // x=level, y=layer, z=layout
    #endif

    //
    // Generated sampling functions:
    //   > SAMPLER_NAME##SampleView(uv)
    //   > SAMPLER_NAME##SampleViewLevel(uv)
    //   > SAMPLER_NAME##SampleViewLevel(uv, level)
    //

    #ifdef FRAGMENT_SHADER
    vec4 SAMPLER_NAME##Sample(vec2 uv) {
        #if SAMPLER_NAME##LayoutRuntime == 0 && SAMPLER_NAME##Layout == sc_TextureRenderingLayout_Regular
        return texture2D(SAMPLER_NAME, uv);
        #else
        return sc_SampleView(SAMPLER_NAME, float(SAMPLER_NAME##Layout), 0.0, uv, SAMPLER_NAME##PixelHeight);
        #endif
    }
    vec4 SAMPLER_NAME##SampleView(vec2 uv, float viewIndex) {
        #if SAMPLER_NAME##LayoutRuntime == 0 && SAMPLER_NAME##Layout == sc_TextureRenderingLayout_Regular
        return texture2D(SAMPLER_NAME, uv);
        #else
        return sc_SampleView(SAMPLER_NAME, float(SAMPLER_NAME##Layout), viewIndex, uv, SAMPLER_NAME##PixelHeight);
        #endif
    }
    // Sample current stereo view.
    vec4 SAMPLER_NAME##SampleView(vec2 uv) {
        float viewIndex = sc_GetStereoViewIndex();
        return SAMPLER_NAME##SampleView(uv, viewIndex);
    }
    #endif // #ifdef FRAGMENT_SHADER

    #if defined(sc_CanUseTextureLod) && (SAMPLER_NAME##SamplerTypeIndex != sc_SamplerTypeIndex_samplerExternalOES)
    vec4 SAMPLER_NAME##SampleLevel(vec2 uv, float level) {
        #if SAMPLER_NAME##LayoutRuntime == 0 && SAMPLER_NAME##Layout == sc_TextureRenderingLayout_Regular
        return texture2DLod(SAMPLER_NAME, uv, level);
        #else
        return sc_SampleViewLevel(SAMPLER_NAME, float(SAMPLER_NAME##Layout), 0.0, uv, SAMPLER_NAME##PixelHeight, level);
        #endif
    }
    vec4 SAMPLER_NAME##SampleViewLevel(vec2 uv, float level) {
        #if SAMPLER_NAME##LayoutRuntime == 0 && SAMPLER_NAME##Layout == sc_TextureRenderingLayout_Regular
        return texture2DLod(SAMPLER_NAME, uv, level);
        #else
        float viewIndex = sc_GetStereoViewIndex();
        return sc_SampleViewLevel(SAMPLER_NAME, float(SAMPLER_NAME##Layout), viewIndex, uv, SAMPLER_NAME##PixelHeight, level);
        #endif
    }
    // Sample current level.
    vec4 SAMPLER_NAME##SampleViewLevel(vec2 uv) {
        return SAMPLER_NAME##SampleViewLevel(uv, SAMPLER_NAME##Level);
    }
    #endif

#endmacro // #macro DECLARE_TEXTURE_SAMPLER_UNIFORMS(SAMPLER_TYPE, SAMPLER_NAME)

// Macros:
//   > sc_ScreenTextureLayout
//   > sc_ScreenTextureSamplerType
//   > sc_ScreenTexture

DECLARE_TEXTURE_SAMPLER_UNIFORMS(sc_ScreenTexture)

#ifdef FRAGMENT_SHADER
vec4 sc_SampleScreenTexture(vec2 uv) {
    return sc_ScreenTextureSample(uv);
}
vec4 sc_SampleMapScreenTexture(vec2 uv) {
    return sc_ScreenTextureSampleView(uv);
}
#else // #ifdef FRAGMENT_SHADER
#ifdef sc_CanUseTextureLod
vec4 sc_SampleMapScreenTexture(vec2 uv) {
    return sc_ScreenTextureSampleViewLevel(uv, 0.0);
}
#endif
#endif // #else // #ifdef FRAGMENT_SHADER

#ifdef sc_CanUseTextureLod
vec4 sc_SampleMapScreenTexture(vec2 uv, float level) {
    return sc_ScreenTextureSampleViewLevel(uv, level);
}
#endif

// A more comprehensive alias for lens shaders.
#ifndef sc_SampleScreenTextureView
#define sc_SampleScreenTextureView sc_SampleMapScreenTexture
#endif // #ifndef sc_SampleScreenTextureView

// Deprecated, use sc_SampleScreenTextureView().
#ifndef sc_SampleScreenTextureForCurrentEye
#define sc_SampleScreenTextureForCurrentEye sc_SampleScreenTextureView
#endif // #ifndef sc_SampleScreenTextureForCurrentEye

#if sc_ExporterVersion < 74
#include <std_shadows.glsl>  // HACK 05/15/2018: some particle shaders were shipped in public studio that did not directly include this, but relied on an old hack that included std_shadows.glsl in std.glsl. Once all those old assets die, we can remove this. Remove related hack from std_shadows.glsl.
#endif
